package ru.dbotthepony.mc.otm.block.entity.worker;

import net.minecraft.core.BlockPos;
import net.minecraft.nbt.CompoundTag;
import net.minecraft.nbt.DoubleTag;
import net.minecraft.world.level.Level;
import net.minecraft.world.level.block.Block;
import net.minecraft.world.level.block.entity.BlockEntity;
import net.minecraft.world.level.block.entity.BlockEntityType;
import net.minecraft.world.level.block.state.BlockState;
import ru.dbotthepony.mc.otm.block.entity.BlockEntityMatteryPowered;
import ru.dbotthepony.mc.otm.capability.MatteryCapability;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.math.BigDecimal;

abstract public class BlockEntityMatteryWorker extends BlockEntityMatteryPowered {
	public BlockEntityMatteryWorker(BlockEntityType<?> p_155228_, BlockPos p_155229_, BlockState p_155230_) {
		super(p_155228_, p_155229_, p_155230_);
	}

	@Nonnull
	protected abstract BigDecimal getBaseConsumption();

	@Nonnull
	protected BigDecimal consumptionPerWork() {
		return getBaseConsumption();
	}

	protected double work_ticks = 0;
	protected int throttle_ticks = 0;
	protected MachineJob current_job;

	// if is_idling is true, then workerLoop() does nothing
	protected boolean is_idling = false;

	@Nullable
	public MachineJob getCurrentJob() {
		return current_job;
	}

	public boolean cantProcessJob() {
		return throttle_ticks > 0;
	}

	@Override
	public CompoundTag save(CompoundTag nbt) {
		nbt.putDouble("work_ticks", work_ticks);

		if (current_job != null) {
			nbt.put("current_job", current_job.serializeNBT());
		}

		return super.save(nbt);
	}

	@Override
	public void load(CompoundTag nbt) {
		super.load(nbt);

		if (nbt.get("work_ticks") instanceof DoubleTag tag)
			work_ticks = tag.getAsDouble();

		current_job = MachineJob.deserializeNBT(nbt.get("current_job"));

		if (current_job == null)
			work_ticks = 0d;
	}

	@Override
	public void setChanged() {
		super.setChanged();
		is_idling = false;
	}

	/**
	 * @param job current job
	 * @return whenever machine can finish it's job. return false if machine for whatever reason can't finish it's job,
	 *         waiting on conditions to be met
	 */
	@Nonnull
	abstract protected MachineJobStatus onJobFinish(MachineJob job);

	/**
	 * @param context context for current job
	 * @return whenever machine can perform it
	 */
	protected MachineJobStatus onWorkTick(WorkTickContext context) {
		return new MachineJobStatus();
	}

	private int idle_ticks_anim = 0;
	private int working_ticks_anim = 0;
	private int error_ticks_anim = 0;

	protected void workerLoop() {
		if (level != null && error_ticks_anim > 20 && getBlockState().hasProperty(WorkerState.WORKER_STATE) && getBlockState().getValue(WorkerState.WORKER_STATE) != WorkerState.ERROR) {
			level.setBlock(getBlockPos(), getBlockState().setValue(WorkerState.WORKER_STATE, WorkerState.ERROR), Block.UPDATE_CLIENTS);
		}

		if (throttle_ticks > 0) {
			working_ticks_anim = 0;
			idle_ticks_anim = 0;
			throttle_ticks--;
			error_ticks_anim++;

			if (throttle_ticks > 0)
				return;
		}

		if (is_idling) {
			working_ticks_anim = 0;
			error_ticks_anim = 0;
			idle_ticks_anim++;

			if (level != null && idle_ticks_anim > 20 && getBlockState().hasProperty(WorkerState.WORKER_STATE) && getBlockState().getValue(WorkerState.WORKER_STATE) != WorkerState.IDLE) {
				level.setBlock(getBlockPos(), getBlockState().setValue(WorkerState.WORKER_STATE, WorkerState.IDLE), Block.UPDATE_CLIENTS);
			}

			return;
		}

		if (current_job == null) {
			MachineJob input = getNextJob();

			if (input == null) {
				is_idling = true;
				working_ticks_anim = 0;
				return;
			}

			current_job = input;
		}

		if (current_job.power_consumption_multiplier().compareTo(BigDecimal.ZERO) != 0 && energy.getBatteryLevel().compareTo(BigDecimal.ZERO) == 0) {
			idle_ticks_anim++;

			if (level != null && idle_ticks_anim > 20 && getBlockState().hasProperty(WorkerState.WORKER_STATE) && getBlockState().getValue(WorkerState.WORKER_STATE) != WorkerState.IDLE) {
				level.setBlock(getBlockPos(), getBlockState().setValue(WorkerState.WORKER_STATE, WorkerState.IDLE), Block.UPDATE_CLIENTS);
			}

			return;
		}

		idle_ticks_anim = 0;

		if (work_ticks < current_job.ticks_processing_time()) {
			if (current_job.power_consumption_multiplier().compareTo(BigDecimal.ZERO) != 0) {
				BigDecimal required_power = consumptionPerWork().multiply(current_job.power_consumption_multiplier());
				BigDecimal extracted_power = energy.extractEnergyInner(required_power, true);

				BigDecimal work_speed = extracted_power.divide(required_power, MatteryCapability.ROUND_RULES);

				MachineJobStatus status = onWorkTick(new WorkTickContext(current_job, required_power, extracted_power, work_speed));

				if (!status.valid()) {
					throttle_ticks += status.throttle();
					return;
				}

				working_ticks_anim++;
				error_ticks_anim = 0;

				double new_work_ticks = work_speed.doubleValue() + work_ticks;

				if (new_work_ticks > current_job.ticks_processing_time()) {
					work_ticks = current_job.ticks_processing_time();
					energy.extractEnergyInner(extracted_power.multiply(new BigDecimal(1d - (new_work_ticks - current_job.ticks_processing_time()), MatteryCapability.ROUND_RULES)), false);
				} else {
					work_ticks = new_work_ticks;
					energy.extractEnergyInner(extracted_power, false);

					if (work_ticks >= current_job.ticks_processing_time()) {
						MachineJobStatus finish = onJobFinish(current_job);

						if (finish.valid()) {
							current_job = null;
							work_ticks = 0d;
						} else {
							throttle_ticks += finish.throttle();
						}
					}
				}
			} else {
				MachineJobStatus status = onWorkTick(new WorkTickContext(current_job, BigDecimal.ZERO, BigDecimal.ZERO, BigDecimal.ONE));

				if (!status.valid()) {
					throttle_ticks += status.throttle();
					return;
				}

				working_ticks_anim++;
				error_ticks_anim = 0;
				work_ticks += 1d;

				if (work_ticks >= current_job.ticks_processing_time()) {
					MachineJobStatus finish = onJobFinish(current_job);

					if (finish.valid()) {
						current_job = null;
						work_ticks = 0d;
					} else {
						throttle_ticks += finish.throttle();
					}
				}
			}
		} else {
			MachineJobStatus finish = onJobFinish(current_job);

			if (finish.valid()) {
				current_job = null;
				work_ticks = 0d;
				error_ticks_anim = 0;
			} else {
				throttle_ticks += finish.throttle();
				error_ticks_anim++;
			}
		}

		if (level != null && (working_ticks_anim > 20 && error_ticks_anim == 0) && getBlockState().hasProperty(WorkerState.WORKER_STATE) && getBlockState().getValue(WorkerState.WORKER_STATE) != WorkerState.WORKING) {
			level.setBlock(getBlockPos(), getBlockState().setValue(WorkerState.WORKER_STATE, WorkerState.WORKING), Block.UPDATE_CLIENTS);
		}
	}

	public double getWorkTicks() {
		return work_ticks;
	}

	public double getWorkProgress() {
		if (current_job == null)
			return 0d;

		return Math.min(1d, work_ticks / current_job.ticks_processing_time());
	}

	/**
	 * Determine which item can be processed from input slots if idling
	 * @return any item in input slots. null if no work is available
	 */
	@Nullable
	protected abstract MachineJob getNextJob();

	public static <T extends BlockEntity> void basicTicker(Level level, BlockPos blockPos, BlockState blockState, T t) {
		if (t instanceof BlockEntityMatteryWorker tile) {
			tile.batteryChargeLoop();
			tile.workerLoop();
		}
	}
}
