package ru.dbotthepony.mc.otm.capability.drive;

import net.minecraft.nbt.CompoundTag;
import net.minecraft.nbt.ListTag;
import net.minecraft.nbt.Tag;
import net.minecraft.resources.ResourceLocation;
import net.minecraft.world.item.Item;
import net.minecraft.world.item.ItemStack;
import net.minecraft.world.item.Items;
import net.minecraftforge.registries.RegistryManager;
import ru.dbotthepony.mc.otm.OverdriveThatMatters;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.*;

public class MatteryDrive implements IMatteryDrive {
	protected final HashMap<Item, List<StoredStack>> items = new HashMap<>();
	protected final HashMap<UUID, StoredStack> items_by_id = new HashMap<>();

	protected boolean dirty = false;
	protected int different_stacks = 0;
	protected int stored = 0;
	protected int max_different_stacks;
	protected int capacity;

	public MatteryDrive(int capacity, int max_different_stacks) {
		this.capacity = capacity;
		this.max_different_stacks = max_different_stacks;
	}

	public MatteryDrive(int capacity) {
		this(capacity, 0xFFFF);
	}

	@Override
	public List<StoredStack> getItems() {
		int amount = 0;

		for (var list : items.values()) {
			amount += list.size();
		}

		var output = new ArrayList<StoredStack>(amount);

		for (var list : items.values()) {
			output.addAll(list);
		}

		return output;
	}

	@Override
	public boolean isDirty() {
		return dirty;
	}

	@Override
	public void markDirty() {
		dirty = true;
	}

	@Override
	public void markClean() {
		dirty = false;
	}

	@Override
	public int getStoredCount() {
		return stored;
	}

	@Override
	public int getCapacity() {
		return capacity;
	}

	@Override
	@Nonnull
	public List<StoredStack> findItems(@Nonnull Item item) {
		var list = items.get(item);

		if (list != null)
			return List.copyOf(list);

		return List.of();
	}

	@Override
	@Nonnull
	public List<StoredStack> findItems(@Nonnull ItemStack stack) {
		var list = items.get(stack.getItem());

		if (list == null)
			return List.of();

		var amount = 0;

		for (var _stack : list) {
			if (ItemStack.tagMatches(_stack.stack(), stack)) {
				amount++;
			}
		}

		var build_list = new ArrayList<StoredStack>(amount);
		var i = 0;

		for (var _stack : list) {
			if (ItemStack.tagMatches(_stack.stack(), stack)) {
				build_list.set(i, _stack);
				i++;
			}
		}

		return build_list;
	}

	protected final ArrayList<IItemViewListener> listeners = new ArrayList<>();

	@Override
	public void addListener(IItemViewListener listener) {
		if (listeners.contains(listener))
			return;

		listeners.add(listener);

		for (var list : this.items.values()) {
			for (var state : list) {
				listener.addViewItem(state.stack(), state.id());
			}
		}
	}

	@Override
	public void removeListener(IItemViewListener listener) {
		if (listeners.remove(listener)) {
			for (var list : this.items.values()) {
				for (var state : list) {
					listener.removeViewItem(state.id());
				}
			}
		}
	}

	@Nonnull
	@Override
	public ItemStack getItem(UUID id) {
		var stack = items_by_id.get(id);
		return stack != null ? stack.stack() : ItemStack.EMPTY;
	}

	@Nonnull
	@Override
	public ItemStack insertItem(@Nonnull ItemStack item, boolean simulate) {
		int max_insert = Math.min(getCapacity() - getStoredCount(), item.getCount());

		if (max_insert <= 0)
			return item;

		var listing = items.computeIfAbsent(item.getItem(), (key) -> new ArrayList<>());

		for (var state : listing) {
			if (ItemStack.tagMatches(state.stack(), item)) {
				if (!simulate) {
					state.stack().grow(max_insert);
					stored += max_insert;

					for (var listener : listeners) {
						listener.changeViewItem(state.id(), state.stack().getCount());
					}

					markDirty();
				}

				var copy_item = item.copy();
				copy_item.shrink(max_insert);
				return copy_item;
			}
		}

		if (different_stacks >= max_different_stacks) {
			return item;
		}

		if (!simulate) {
			different_stacks++;
			stored += max_insert;

			var copy = item.copy();
			copy.setCount(max_insert);
			var state = new StoredStack(copy, UUID.randomUUID());
			listing.add(state);
			items_by_id.put(state.id(), state);

			for (var listener : listeners) {
				listener.addViewItem(state.stack(), state.id());
			}

			markDirty();
		}

		var copy_item = item.copy();
		copy_item.shrink(max_insert);
		return copy_item;
	}

	@Nonnull
	@Override
	public ItemStack extractItem(UUID id, int amount, boolean obey_stack_size, boolean simulate) {
		var get = items_by_id.get(id);

		if (get == null)
			return ItemStack.EMPTY;

		var extract = Math.min(amount, get.stack().getCount());

		if (obey_stack_size)
			extract = Math.min(extract, get.stack().getItem().getItemStackLimit(get.stack()));

		if (extract <= 0)
			return ItemStack.EMPTY;

		var copy = get.stack().copy();
		copy.setCount(extract);

		if (!simulate) {
			if (extract == get.stack().getCount()) {
				var listing = items.get(get.stack().getItem());
				listing.remove(get);
				different_stacks--;

				for (var listener : listeners) {
					listener.removeViewItem(get.id());
				}

				if (listing.size() == 0) {
					items.remove(get.stack().getItem());
				}
			}

			stored -= extract;
			get.stack().shrink(extract);

			if (get.stack().getCount() != 0) {
				for (var listener : listeners) {
					listener.changeViewItem(get.id(), get.stack().getCount());
				}
			}

			markDirty();
		}

		return copy;
	}

	@Override
	public CompoundTag serializeNBT() {
		final var compound = new CompoundTag();

		compound.putInt("capacity", capacity);
		compound.putInt("max_different_stacks", max_different_stacks);

		var list_items = new ListTag();
		compound.put("items", list_items);

		for (var entry : items.entrySet()) {
			var item = Objects.requireNonNull(entry.getKey().getRegistryName(), "Missing registry name for stored Item").toString();
			var entry_compound = new CompoundTag();

			entry_compound.putString("id", item);
			var stack_list = new ListTag();
			entry_compound.put("list", stack_list);

			for (var stack : entry.getValue()) {
				var stack_nbt = new CompoundTag();
				stack_list.add(stack_nbt);

				stack_nbt.putInt("count", stack.stack().getCount());
				stack_nbt.putLongArray("id", new long[] { stack.id().getMostSignificantBits(), stack.id().getLeastSignificantBits() });

				if (stack.stack().getTag() != null) {
					stack_nbt.put("data", stack.stack().getTag());
				}
			}

			list_items.add(entry_compound);
		}

		return compound;
	}

	@Override
	public void deserializeNBT(CompoundTag nbt) {
		items.clear();
		items_by_id.clear();
		stored = 0;
		different_stacks = 0;

		capacity = nbt.getInt("capacity");
		max_different_stacks = nbt.getInt("max_different_stacks");

		var list_items = nbt.getList("items", Tag.TAG_COMPOUND);
		var registry = RegistryManager.ACTIVE.getRegistry(Item.class);

		for (var _entry : list_items) {
			if (_entry instanceof CompoundTag entry) {
				var item = registry.getValue(new ResourceLocation(entry.getString("id")));

				if (item != null && item != Items.AIR) {
					var map_list = items.computeIfAbsent(item, (key) -> new ArrayList<>());

					var stack_list = entry.getList("list", Tag.TAG_COMPOUND);

					for (var _stack : stack_list) {
						if (_stack instanceof CompoundTag stack) {
							var count = stack.getInt("count");
							var id = stack.getLongArray("id");
							var data = stack.get("data");

							var itemstack = new ItemStack(item, count);

							if (data != null) {
								itemstack.setTag((CompoundTag) data);
							}

							stored += count;
							different_stacks += 1;
							var state = new StoredStack(itemstack, id.length == 2 ? new UUID(id[0], id[1]) : UUID.randomUUID());
							map_list.add(state);
							items_by_id.put(state.id(), state);
						}
					}
				}
			}
		}
	}
}
